import fetch from 'node-fetch'
import crypto from 'crypto'
import { config } from './config.js'
import { getRedisKey, isJSFile } from './util.js'
import { getCSSDependenciesFor, getViteManifests } from './manifests.js'
import * as cache from './cache.js'
import { logger } from './logger.js'
import { NotFoundError } from './errors.js'

export async function fetchFileWithHeadersFromBaseUrl (path, baseUrl, version) {
  const [response, dependencies] = await Promise.all([
    fetch(new URL(path, baseUrl)),
    isJSFile(path) && getCSSDependenciesFor({ file: path.substr(1), version })
  ])

  if (!response.ok) throw new NotFoundError(`Error fetching file: ${path}`)

  const cssString = dependencies && dependencies.map(file => `"${file}"`).join(',')
  const appendix = cssString && `\n/*injected by ui-middleware*/document.dispatchEvent(new CustomEvent("load-css",{detail:{css:[${cssString}]}}))`
  const resBuffer = await response.arrayBuffer()
  const appendixLength = appendix?.length || 0
  const body = Buffer.alloc(resBuffer.byteLength + appendixLength)

  body.fill(Buffer.from(resBuffer), 0, resBuffer.byteLength)
  if (appendix) body.write(appendix, resBuffer.byteLength)

  const sha256Sum = crypto.createHash('sha256').update(body).digest('base64')

  return {
    body,
    sha256Sum,
    headers: {
      'content-type': response.headers.get('content-type'),
      dependencies
    }
  }
}

export async function fetchFileWithHeaders ({ path, version }) {
  const viteManifests = await getViteManifests({ version })
  const module = viteManifests[path.substr(1)]
  if (module?.meta?.baseUrl) {
    try {
      return fetchFileWithHeadersFromBaseUrl(path, module.meta.baseUrl, version)
    } catch (err) {
      logger.debug(`[Files] File ${path} had a baseUrl but could not be found on that server: ${err}`)
    }
  }

  return Promise.any(config.urls.map(baseUrl => fetchFileWithHeadersFromBaseUrl(path, baseUrl, version)))
}

export async function saveAsyncToCache ({ version, path, promise }) {
  await Promise.all([
    cache.setAsync(getRedisKey({ version, name: `${path}:body` }), promise.then(({ body }) => {
      if (typeof body !== 'string' && !(body instanceof Buffer)) return JSON.stringify(body)
      return body
    })),
    cache.setAsync(getRedisKey({ version, name: `${path}:meta` }), promise.then(({ body, ...rest }) => JSON.stringify(rest)))
  ])
  return promise
}

export async function loadFromCache ({ version, path }) {
  const [body, meta = '{}'] = await Promise.all([
    cache.getBuffer(getRedisKey({ version, name: `${path}:body` })),
    cache.get(getRedisKey({ version, name: `${path}:meta` }))
  ])
  if (!body) return
  return { ...JSON.parse(meta), body }
}
